from random import randint, random, choice
from math import sin, tan, radians, copysign, degrees, cos, asin
from os.path import join

from pygame.locals import *
from pygame import Surface, Color, PixelArray
from pygame.font import Font
from pygame.mixer import Sound, Channel, get_num_channels
from pygame.draw import polygon, line, circle
from pygame.gfxdraw import aapolygon, aacircle, filled_circle
from pygame.image import load
from pygame.transform import rotate, smoothscale, rotozoom

from _lib.pgfw.pgfw.Game import Game
from _lib.pgfw.pgfw.GameChild import GameChild
from _lib.pgfw.pgfw.Sprite import Sprite
from _lib.pgfw.pgfw.Animation import Animation
from _lib.pgfw.pgfw.extension import get_distance

class SoundEffect(GameChild, Sound):

    def __init__(self, parent, path, volume):
        GameChild.__init__(self, parent)
        Sound.__init__(self, path)
        self.display_surface = self.get_display_surface()
        self.set_volume(volume)

    def play(self, loops=0, maxtime=0, fade_ms=0, position=None, x=None):
        channel = Sound.play(self, loops, maxtime, fade_ms)
        if x is not None:
            position = float(x) / self.display_surface.get_width()
	if position is not None and channel is not None:
            channel.set_volume(*self.get_panning(position))
        return channel

    def get_panning(self, position):
        return 1 - max(0, ((position - .5) * 2)), \
               1 + min(0, ((position - .5) * 2))


class Vector(GameChild):

    def __init__(self, parent, x, y, z):
        GameChild.__init__(self, parent)
        self.x, self.y, self.z = x, y, z

    def __repr__(self):
        return "<%.3f %.3f %.3f>" % (self.x, self.y, self.z)

    def rotate(self, th, x=0, y=0, z=0):
        k = x, y, z
        v = self.x, self.y, self.z
        c = cos(th)
        vs = [cc * c for cc in v]
        s = sin(th)
        vc = [(k[1] * v[2] - k[2] * v[1]) * s,
              (k[2] * v[0] - k[0] * v[2]) * s,
              (k[0] * v[1] - k[1] * v[0]) * s]
        dp = sum(k[ii] * v[ii] for ii in xrange(len(k)))
        kd = [cc * dp * (1 - c) for cc in k]
        self.x, self.y, self.z = (vs[ii] + vc[ii] + kd[ii] for ii in \
                                  xrange(len(v)))

    def get_screen_coordinates(self, radius, center=None):
        if center is None:
            center = self.display_surface.get_rect().center
        cx, cy = center
        x = int(round(cx + self.x * radius))
        y = int(round(cy + self.y * radius))
        return x, y


class Wisp(Game, Animation):

    LEVEL_DATA = (((213, 255, 34, 200), (134, 91, 42, 128),
                   (220, 255), (190, 240), (190, 240),
                   (192, 87, 30), (0, 180, 0),
                   (((-3, -3, 2), (3, 1, -2), (0, 0, 4), (100, 100, 200), (30, (450, 200))),
                    ((-3, -3, 2), (0, 0, 4), (-1, -5, 0), (80, 100, 155), (30, (450, 200))),
                    ((-1, 5, 0), (0, 0, 4), (1, 1, -2), (122, 68, 72), (30, (450, 200))),
                    ((1, 1, -2), (-1, 5, 0), (-2, 4, 3), (180, 60, 180), (20, (450, 300)))), 1),

                  ((235, 24, 82, 130), (86, 208, 64, 200),
                   (180, 210), (190, 240), (190, 240),
                   (213, 255, 34), (240, 20, 200),
                   (((5, 2, -1), (1, -3, 2), (-3, 5, 3), (222, 122, 0), (5, (450, 210))),
                    ((1, -3, 2), (-3, 5, 3), (4, 1, 8), (30, 153, 87), (5, (450, 210))),
                    ((-3, 5, 3), (4, 1, 8), (2, -3, -3), (200, 130, 10), (5, (450, 210))),
                    ((2, 4, 2), (8, 11, -4), (-14, -4, 3), (30, 180, 90), (11, (450, 290)))), 2),
                  
                  ((100, 100, 255, 160), (82, 64, 123, 240),
                   (220, 250), (220, 250), (120, 180),
                   (0, 255, 255), (50, 40, 30),
                   (((2, 1, 2), (4, -2, -2), (-5, 3, -2), (200, 140, 201), (35, (450, 190))),
                    ((2, 1, 2), (-5, 3, -2), (-1, 6, 1), (30, 87, 153), (35, (450, 190))),
                    ((-5, 3, -2), (-1, 0, 2), (3, 5, -1), (30, 87, 153), (20, (450, 300))),
                    ((-1, 0, 2), (3, 5, -1), (1, 10, -3), (255, 142, 193), (20, (450, 300)))), 3),

                  ((42, 91, 134, 200), (255, 34, 213, 128),
                   (200, 255), (190, 200), (200, 255),
                   (78, 98, 28), (213, 255, 34),
                   (((-3, 4, 1), (0, 1, 0), (-2, -2, -1), (200, 200, 100), (22, (450, 200))),
                    ((-3, 4, 1), (0, 1, 0), (1, 5, 3), (140, 140, 80), (22, (450, 200))),
                    ((-2, -2, -1), (1, 5, 3), (-1, -1, -3), (85, 85, 85), (22, (450, 200))),
                    ((0, -3, 2), (3, 4, -2), (-2, -1, -1), (202, 101, 90), (13, (450, 240))),
                    ((3, 4, -2), (-2, -1, -1), (2, 0, 1), (140, 140, 80), (13, (450, 280)))), 4),

                  ((100, 255, 100, 80), (64, 143, 82, 170),
                   (210, 240), (240, 255), (180, 200),
                   (60, 60, 120), (200, 120, 120),
                   (((-2, 5, 3), (3, 1, -3), (4, -2, 1), (48, 84, 148), (30, (450, 180))),
                    ((3, 1, -3), (4, -2, 1), (1, 1, 0), (32, 200, 200), (30, (450, 180))),
                    ((1, 1, 0), (4, -2, 1), (-3, -2, -1), (48, 148, 84), (30, (450, 180)),
                    (-3, -4, 1), (3, -4, 1), (0, 4, 2), (200, 30, 120), (25, (450, 300))),
                    ((0, 4, 2), (3, -4, 1), (2, 0, -4), (48, 148, 84), (25, (450, 300)))), 5),

                  ((32, 157, 85, 230), (200, 100, 100, 170),
                   (180, 230), (130, 180), (240, 255),
                   (80, 80, 80), (255, 255, 255),
                   (((3, 2, 1), (1, 2, 3), (-2, -3, -1), (52, 103, 183), (22, (430, 160))),
                    ((1, 2, 3), (-2, -3, -1), (-5, -3, 4), (61, 133, 172), (22, (430, 160))),
                    ((-5, -3, 4), (-2, -3, -1), (2, 1, 0), (88, 109, 255), (22, (430, 160))),
                    ((3, 2, 1), (2, 1, 0), (-5, -3, 4), (92, 255, 0), (22, (430, 160))),
                    ((5, 8, 2), (3, -3, -5), (-2, -1, 3), (222, 222, 102), (22, (430, 240))),
                    ((2, 1, 5), (10, 1, 7), (-3, -4, -2), (30, 30, 180), (22, (430, 240)))), 6))

    FORM_ROTATION_SPEED = .01
    FORM_ROTATION_LIMIT = .25
    IN_GAME_MUSIC_FADE_STEP = .01
    
    def __init__(self):
        Game.__init__(self)
        Animation.__init__(self, self)
        self.platform = Platform(self)
        self.message = Message(self)
        self.drop = Drop(self)
        self.best = Best(self)
        self.title_music = Sound(self.get_resource("LLll75E.ogg"))
        self.title_music.set_volume(.75)
        self.in_game_music = Sound(self.get_resource("BOGSWEATer.ogg"))
        self.end_music = Sound(self.get_resource("Hey_Who_is_This.ogg"))
        self.end_music.set_volume(.5)
        self.set_levels()
        self.subscribe(self.respond)
        self.register(self.unsuppress_commands, self.fade_out_in_game_music)
        self.reset()

    def print_channel_contents(self):
        print "channels\n----------"
        for index in xrange(get_num_channels()):
            channel = Channel(index)
            sound = channel.get_sound()
            print channel.get_busy(), sound, sound.get_length() if sound else None
        print

    def set_levels(self):
        levels = self.levels = []
        for ii, parameters in enumerate(self.LEVEL_DATA):
            levels.append(Level(self, ii, parameters[:2], parameters[2:5], parameters[5:7], parameters[7], parameters[8]))

    def respond(self, event):
        if not self.suppressing_commands:
            if self.delegate.compare(event, "reset-game"):
                self.reset()
            elif self.delegate.compare(event, "any") and self.waiting_to_start:
                self.message.set_text("stop near the horizon")
                self.in_game_music_channel = self.in_game_music.play(-1, 0, 3000)
                self.title_music.fadeout(3000)
                self.waiting_to_start = False
                self.in_game = True
                self.drop.begin_level()
            elif self.delegate.compare(event, "any") and self.game_complete:
                self.reset()
                self.game_complete = False

    def unsuppress_commands(self):
        self.suppressing_commands = False
        
    def fade_out_in_game_music(self):
        pass

    def reset(self):
        self.level_index = 0
        self.form_rotation = 0
        self.cumulative_time = 0
        self.form_rotation_direction = choice((-1, 1))
        self.platform.reset()
        self.message.unhighlight()
        self.message.set_text("press any key to begin")
        self.title_music.fadeout(3000)
        if self.in_game_music.get_num_channels() or self.end_music.get_num_channels():
            self.title_music.play(-1, 0, 3000)
        else:
            self.title_music.play(-1)
        self.game_complete = False
        self.suppressing_commands = True
        self.play(self.unsuppress_commands, delay=1000, play_once=True)
        self.waiting_to_start = True
        self.in_game = False
        self.in_game_music.fadeout(3000)
        self.end_music.fadeout(3000)
        self.drop.reset()

    def get_level(self):
        return self.levels[self.level_index]

    def advance_level(self):
        self.level_index += 1
        if self.level_index == len(self.LEVEL_DATA):
            self.level_index = len(self.LEVEL_DATA) - 1
            self.game_complete = True
            self.message.highlight()
            self.message.set_text("time: %.1fs" % (self.cumulative_time / 1000.0))
            self.suppressing_commands = True
            self.play(self.unsuppress_commands, delay=1000, play_once=True)
            self.in_game_music.fadeout(3000)
            self.end_music.play(-1, 0, 3000)
            self.drop.active = False
            self.best.add(self.cumulative_time)
        else:
            self.drop.begin_level()

    def update(self):
        Animation.update(self)
        self.get_screen().fill([randint(*limits) for limits in self.get_level().background])
        if not self.game_complete:
            self.drop.update()
        self.platform.update()
        self.platform.update_back_trees()
        if not self.game_complete:
            for shape in self.get_level().form:
                points = []
                for vector in shape[:3]:
                    vector.rotate(self.form_rotation_direction * self.FORM_ROTATION_SPEED, x=1)
                    points.append(vector.get_screen_coordinates(*shape[4]))
                aapolygon(self.get_screen(), points, shape[3])
                polygon(self.get_screen(), shape[3], points)
            self.form_rotation += self.form_rotation_direction * self.FORM_ROTATION_SPEED
            if self.form_rotation < -self.FORM_ROTATION_LIMIT:
                self.form_rotation += self.FORM_ROTATION_SPEED * 2
                self.form_rotation_direction = -self.form_rotation_direction
            elif self.form_rotation > self.FORM_ROTATION_LIMIT:
                self.form_rotation -= self.FORM_ROTATION_SPEED * 2
                self.form_rotation_direction = -self.form_rotation_direction
        self.platform.update_front_trees()
        self.message.update()
        self.best.update()


class Message(Sprite):

    def __init__(self, parent):
        Sprite.__init__(self, parent)

    def set_text(self, text):
        self.clear_frames()
        color = Color(0, 0, 0)
        spaced = text[0]
        for char in text[1:]:
            spaced += " " + char
        for hue in xrange(0, 360, 10):
            color.hsla = hue, 100, 40, 100
            font = Font(self.get_resource("AlegreyaSansSC-Thin.ttf"), (28, 68)[self.highlit])
            self.add_frame(font.render(spaced, True, color))
        ds = self.get_display_surface().get_rect()
        if self.highlit:
            self.location.center = ds.center
        else:
            self.location.center = ds.centerx, ds.h - 25

    def clear_text(self):
        self.clear_frames()

    def highlight(self):
        self.highlit = True

    def unhighlight(self):
        self.highlit = False


class Level(GameChild):

    def __init__(self, parent, index, platform_colors, background, tree_colors, form_coordinates, fall_speed):
        GameChild.__init__(self, parent)
        self.index = index
        self.platform_colors = platform_colors
        self.background = background
        self.tree_colors = tree_colors
        form = self.form = []
        for shape in form_coordinates:
            form.append([Vector(self, *coordinates) for coordinates in shape[:3]] + list(shape[3:]))
        self.fall_speed = fall_speed
        self.load_ellipse()

    def load_ellipse(self):
        self.ellipse = load(self.get_resource(join("u_IO_C", "%i.png" % self.index))).convert_alpha()


class Platform(Animation):

    ANGLE = 30
    MARGIN = 50, 50
    LENGTH = 50
    BUFFER = 4
    CELL_COUNT = 10

    def __init__(self, parent):
        Animation.__init__(self, parent, self.move)
        length = self.LENGTH
        dx = length / tan(radians(self.ANGLE))
        inset, offset = self.MARGIN
        ds = self.get_display_surface().get_rect()
        surface = self.surface = Surface((ds.w - inset * 2, length), SRCALPHA)
        rect = self.rect = surface.get_rect()
        rect.bottomleft = inset, ds.h - offset
        self.corners = (0, rect.h), (rect.w, rect.h), (rect.w - dx, 0), (dx, 0)
        self.dy_nodeset = self.get_game().interpolator.get_nodeset("platform-dy")
        self.play()

    def move(self):
        self.y_offset -= .5
        if self.y_offset < -self.BUFFER:
            self.y_offset = 0
            self.y_switch = not self.y_switch

    def reset(self):
        self.y_offset = 0
        self.y_switch = False
        self.trees = []
        
    def update(self):
        if random() < .08:
            self.trees.append(Tree(self))
        Animation.update(self)
        background, foreground = self.get_game().get_level().platform_colors
        surface = self.surface.copy()
        aapolygon(surface, self.corners, background)
        polygon(surface, background, self.corners)
        buf = self.BUFFER
        y = self.y_offset
        ys = self.y_switch
        rect = self.rect
        dx = self.corners[-1][0]
        cx = rect.centerx
        while y < rect.h:
            dy = int(self.dy_nodeset.get_y(float(y + buf) / (rect.h + buf)) * (rect.h + buf))
            n = ys
            while True:
                x0 = self.get_x(n)
                n += 1
                x1 = self.get_x(n)
                n += 1
                if x1 > rect.w - dx:
                    break
                x0_a = self.get_angle(x0)
                x1_a = self.get_angle(x1)
                points = (x0 + 1 + copysign(y / tan(x0_a), x0 - cx), y), \
                         (x1 - 1 + copysign(y / tan(x1_a), x1 - cx), y), \
                         (x1 - 1 + copysign((y + dy) / tan(x1_a), x1 - cx), y + dy), \
                         (x0 + 1 + copysign((y + dy) / tan(x0_a), x0 - cx), y + dy)
                aapolygon(surface, points, foreground)
                polygon(surface, foreground, points)
            ys = not ys
            y += dy
        self.get_display_surface().blit(surface, rect)

    def get_x(self, n):
        dx = self.corners[-1][0]
        return int(dx + n * ((self.rect.w - dx * 2) / float(self.CELL_COUNT)))

    def get_angle(self, x):
        return radians(self.ANGLE) * abs(x - self.rect.centerx) / abs(self.corners[-1][0] - self.rect.centerx)
    
    def update_back_trees(self):
        for tree in self.trees:
            if tree.get_y() <= self.rect.h / 2:
                tree.update()

    def update_front_trees(self):
        for tree in self.trees:
            if tree.get_y() > self.rect.h / 2:
                tree.update()


class Tree(Sprite):

    MAX_TRUNK_HEIGHT = 8
    MAX_LEAVES_RADIUS = 38
    MIN_X_SCALE = .4
    STEP_COUNT = 75
    
    def __init__(self, parent):
        Sprite.__init__(self, parent)
        self.xr = random()
        y_steps = self.y_steps = []
        nodeset = self.get_game().interpolator.get_nodeset("tree-y-steps")
        for ii in xrange(self.STEP_COUNT):
            y_steps.append(int(self.parent.rect.h * nodeset.get_y(float(ii) / self.STEP_COUNT)))
        self.yi = 0

    def get_y(self):
        return self.y_steps[self.yi]
    
    def update(self):
        if self.yi >= len(self.y_steps):
            self.parent.trees.remove(self)
        else:
            rect = self.parent.rect
            trunk_c, leaves_c = self.get_game().get_level().tree_colors
            y = self.get_y()
            x_scale = (self.MIN_X_SCALE + (1 - self.MIN_X_SCALE) * (1 - abs(self.xr - .5) * 2))
            y_scale = float(y) / rect.h
            height = int(self.MAX_TRUNK_HEIGHT * x_scale * y_scale) 
            x_in = int((rect.h - y) / tan(radians(self.parent.ANGLE)))
            x = int((rect.w - x_in * 2) * self.xr + x_in)
            radius = int(self.MAX_LEAVES_RADIUS * x_scale * y_scale)
            if radius > 4:
                start = rect.left + x, rect.top + y - 6
                end = rect.left + x, rect.top + y - height - 6
                ds = self.get_display_surface()
                line(ds, trunk_c, start, (end[0], end[1] - 2), max(3, int(round(7 * x_scale * y_scale))))
                aacircle(ds, end[0], end[1] - radius, radius, leaves_c)
                filled_circle(ds, end[0], end[1] - radius, radius, leaves_c)
                Sprite.update(self)
                self.yi += 1


class Drop(Animation):

    BLINK_INTERVAL = 500
    X = 210
    START_Y = 100
    END_MARGIN = 38
    DELAY = 4000
    ACCELERATION = .1
    SAFE_ZONE = 15

    def __init__(self, parent):
        Animation.__init__(self, parent)
        self.limit = self.get_game().platform.rect.top
        self.success_sound = Sound(self.get_resource("man-before-time.ogg"))
        self.success_sound.set_volume(.9)
        self.scale_nodeset = self.get_game().interpolator.get_nodeset("scale")
        self.deceleration_nodeset = self.get_game().interpolator.get_nodeset("deceleration")
        self.subscribe(self.respond)
        self.register(self.blink, interval=self.BLINK_INTERVAL)
        self.register(self.unsuppress_commands, self.drop_bomb)

    def respond(self, event):
        if self.active:
            if not self.suppressing_commands:
                compare = self.get_game().delegate.compare
                if not self.succeeded and not self.failed:
                    if compare(event, "any") and not self.decelerating:
                        self.decelerating = True
                    elif compare(event, "any", True) and self.decelerating:
                        self.decelerating = False
                        self.deceleration_elapsed = 0
                        self.fall_speed = self.get_game().get_level().fall_speed
                elif self.succeeded:
                    self.get_game().advance_level()
                elif self.failed:
                    self.get_game().reset()

    def unsuppress_commands(self):
        self.suppressing_commands = False

    def blink(self):
        self.bomb_visible = not self.bomb_visible

    def reset(self):
        self.active = False
        self.bomb_visible = True
        self.suppressing_commands = False
        self.decelerating = False
        self.waiting_for_end_level_confirm = False
        self.deceleration_elapsed = 0
        self.halt()

    def begin_level(self):
        self.active = True
        self.failed = False
        self.succeeded = False
        self.waiting_for_end_level_confirm = False
        level = self.get_game().get_level()
        if self.get_game().level_index > 0:
            self.get_game().message.clear_text()
        self.bomb_base_image = level.ellipse
        self.bomb_rect = self.bomb_base_image.get_rect()
        self.fall_speed = level.fall_speed
        self.fall_elapsed = 0
        self.suppressing_commands = False
        self.falling = False
        self.neutralized = False
        self.traveled = 0
        self.play(self.blink)
        self.play(self.drop_bomb, delay=self.DELAY, play_once=True)

    def drop_bomb(self):
        self.halt(self.blink)
        self.bomb_visible = True
        self.falling = True

    def update(self):
        if self.active:
            Animation.update(self)
            ds = self.get_display_surface()
            # line(ds, (255, 0, 0), (self.X - 10, self.limit), (self.X + 10, self.limit))
            base = self.bomb_base_image
            if not self.falling and not self.failed and not self.succeeded:
                if self.bomb_visible:
                    rect = base.get_rect()
                    rect.center = self.X, self.START_Y
                    ds.blit(base, rect)
            elif self.falling:
                last_frame_duration = self.get_game().time_filter.get_last_frame_duration()
                self.fall_elapsed += last_frame_duration
                if not self.decelerating:
                    self.fall_speed += self.fall_speed * self.ACCELERATION
                else:
                    self.fall_speed -= self.deceleration_nodeset.get_y(self.deceleration_elapsed)
                    if self.fall_speed < 0: 
                        self.fall_speed = 0
                    self.deceleration_elapsed += last_frame_duration
                self.traveled += self.fall_speed
            if not self.is_playing(self.blink):
                for y in xrange(int(self.traveled), -1, -1):
                    if y == 0:
                        rect = base.get_rect()
                        rect.center = self.X, self.START_Y
                        ds.blit(base, rect)
                    else:
                        scale = .3 + .7 * (1 - y / float(self.limit - self.START_Y))
                        surface = smoothscale(base, (int(base.get_width() * scale), int(base.get_height() * scale)))
                        rect = surface.get_rect()
                        rect.center = self.X, self.START_Y + y
                        ds.blit(surface, rect)
                        if not self.failed and not self.succeeded:
                            if rect.bottom > self.limit:
                                self.failed = True
                                self.falling = False
                            if not self.fall_speed and self.limit - rect.bottom <= self.SAFE_ZONE:
                                self.succeeded = True
                                self.falling = False
                                self.success_sound.play()
            if not self.waiting_for_end_level_confirm:
                if self.failed:
                    self.get_game().message.set_text("game over")
                    self.suppressing_commands = True
                    self.play(self.unsuppress_commands, delay=1000, play_once=True)
                    self.waiting_for_end_level_confirm = True
                elif self.succeeded:
                    self.get_game().message.set_text("%.1f" % (self.fall_elapsed / 1000.0))
                    self.get_game().cumulative_time += self.fall_elapsed
                    self.suppressing_commands = True
                    self.play(self.unsuppress_commands, delay=1000, play_once=True)
                    self.waiting_for_end_level_confirm = True


class Best(Sprite):

    def __init__(self, parent):
        Sprite.__init__(self, parent)
        self.set_low()

    def set_low(self):
        self.clear_frames()
        low = None
        for line in open(self.get_resource("scores")):
            if low is None or int(line) < low:
                low = int(line)
        font = Font(self.get_resource("AlegreyaSansSC-Thin.ttf"), 16)
        for hue in xrange(0, 360, 10):
            color = Color(0, 0, 0)
            color.hsla = hue, 100, 40, 100
            self.add_frame(font.render("hi: %.1fs" % (low / 1000.0), True, color))
        self.location.topright = self.get_display_surface().get_rect().topright
        self.location.right -= 5
        
    def add(self, time):
        fp = open(self.get_resource("scores"), "a")
        fp.write(str(time) + "\n")
        fp.close()
        self.set_low()
